// maunium-stickerpicker - A fast and simple Matrix sticker picker widget.
// Copyright (C) 2020 Tulir Asokan
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.
import {html, render, Component} from "../lib/htm/preact.js"
import {Spinner} from "./spinner.js"
import {SearchBox} from "./search-box.js"
import {giphyIsEnabled, GiphySearchTab, setGiphyAPIKey} from "./giphy.js"
import * as widgetAPI from "./widget-api.js"
import * as frequent from "./frequently-used.js"
import * as userSettings from "./user-settings.js"

// The base URL for fetching packs. The app will first fetch ${PACK_BASE_URL}/index.json,
// then ${PACK_BASE_URL}/${packFile} for each packFile in the packs object of the index.json file.
const PACKS_BASE_URL = "packs"

let INDEX = `${PACKS_BASE_URL}/index.json`
const params = new URLSearchParams(document.location.search)
if (params.has('config')) {
	INDEX = params.get("config")
}

// This is updated from packs/index.json
let HOMESERVER_URL = "https://matrix-client.matrix.org"
let HOMESERVER_ANIMATED_URL = HOMESERVER_URL

const isAnimated = (content) => {
	return content.info["mimetype"] === "image/gif";
};

const makeThumbnailURL = (content) => {
	const homeServerUrl = isAnimated(content) ? HOMESERVER_ANIMATED_URL : HOMESERVER_URL;
	return `${homeServerUrl}/_matrix/media/v3/thumbnail/${content.url.slice(6)}?height=128&width=128&method=scale`;
};

// We need to detect iOS webkit because it has a bug related to scrolling non-fixed divs
// This is also used to fix scrolling to sections on Element iOS
const isMobileSafari = navigator.userAgent.match(/(iPod|iPhone|iPad)/) && navigator.userAgent.match(/AppleWebKit/)

const supportedThemes = ["light", "dark", "black"]

const defaultState = {
	viewingGifs: false,
	packs: [],
	loading: true,
	error: null,
}

const defaultSearchState = {
	searchTerm: null,
	filteredPacks: null
}

const makeFrequentlyUsedState = ({ stickerIDs, stickers } = {}) => ({
	id: "frequently-used",
	title: "Frequently used",
	stickerIDs: stickerIDs ?? [],
	stickers: stickers ?? [],
})

const sanitizeString = s => s.toLowerCase().trim()

class App extends Component {
	constructor(props) {
		super(props)
		this.defaultTheme = params.get("theme")
		this.state = {
			...defaultState,
			...defaultSearchState,
			stickersPerRow: parseInt(localStorage.mauStickersPerRow || "4"),
			theme: localStorage.mauStickerThemeOverride || this.defaultTheme,
			frequentlyUsed: makeFrequentlyUsedState(),
			userSettings: {
				hideGifIndicator: userSettings.shouldHideGifIndicator(),
			},
		}

		if (!supportedThemes.includes(this.state.theme)) {
			this.state.theme = "light"
		}
		if (!supportedThemes.includes(this.defaultTheme)) {
			this.defaultTheme = "light"
		}

		this.imageObserver = null
		this.packListRef = null
		this.navRef = null

		this.searchStickers = this.searchStickers.bind(this)
		this.resetSearch = this.resetSearch.bind(this)
		this.sendSticker = this.sendSticker.bind(this)
		this.navScroll = this.navScroll.bind(this)
		this.reloadPacks = this.reloadPacks.bind(this)
		this.clearFrequentlyUsed = this.clearFrequentlyUsed.bind(this)
		this.observeSectionIntersections = this.observeSectionIntersections.bind(this)
		this.observeImageIntersections = this.observeImageIntersections.bind(this)
	}

	_getStickersByID(ids) {
		return ids.map(id => this.stickersByID.get(id)).filter(sticker => !!sticker)
	}

	_setFrequentlyUsed(stickerIDs = []) {
		const stickers = this._getStickersByID(stickerIDs)
		const frequentlyUsed = makeFrequentlyUsedState({ stickerIDs, stickers })
		this.setState({ frequentlyUsed })
		frequent.setFrequentlyUsedCacheStorage(stickers)
	}

	updateFrequentlyUsed() {
		const stickerIDs = frequent.get()
		this._setFrequentlyUsed(stickerIDs)
	}

	refreshFrequentlyUsed(packs) {
		const stickerIDs = frequent.removeNotFoundFromStorage(packs)
		this._setFrequentlyUsed(stickerIDs)
	}

	clearFrequentlyUsed() {
		frequent.removeAll()
		this._setFrequentlyUsed()
	}

	// Search

	resetSearch() {
		this.setState({ ...defaultSearchState })
	}

	searchStickers(e) {
		const searchTerm = e.target.value
		const sanitizedSearch = sanitizeString(searchTerm)

		const allPacks = [this.state.frequentlyUsed, ...this.state.packs]
		const packsWithFilteredStickers = allPacks.map(pack => ({
			...pack,
			stickers: pack.stickers.filter(sticker =>
				sanitizeString(sticker.body).includes(sanitizedSearch) ||
				sanitizeString(sticker.id).includes(sanitizedSearch)
			),
		}))
		const filteredPacks = packsWithFilteredStickers.filter(({ stickers }) => !!stickers.length)

		this.setState({ searchTerm, filteredPacks })
	}

	// End search

	setStickersPerRow(val) {
		localStorage.mauStickersPerRow = val
		document.documentElement.style.setProperty("--stickers-per-row", localStorage.mauStickersPerRow)
		this.setState({
			stickersPerRow: val,
		})
		this.packListRef.scrollTop = this.packListRef.scrollHeight
	}

	setTheme(theme) {
		if (theme === "default") {
			delete localStorage.mauStickerThemeOverride
			this.setState({theme: this.defaultTheme})
		} else {
			localStorage.mauStickerThemeOverride = theme
			this.setState({theme: theme})
		}
	}

	setDisplayGifIndicator(checked) {
		userSettings.setShouldHideGifIndicator(checked)
		this.setState({
			userSettings: {
				...this.state.userSettings,
				hideGifIndicator: checked,
			},
		})
	}

	// End settings

	reloadPacks() {
		this.imageObserver.disconnect()
		this.sectionObserver.disconnect()
		this.setState({ packs: defaultState.packs })
		this.resetSearch()
		this._loadPacks(true)
	}

	_initializeStickersByID(ids) {
		this.stickersByID = new Map(ids ?? [])
	}

	async populateStickersByID(allPacks) {
		const allStickers = allPacks.map(({ stickers }) => stickers).flat()
		allStickers.forEach((sticker) => {
			this.stickersByID.set(sticker.id, sticker)
		})
	}

	_loadPacks(disableCache = false) {
		const cache = disableCache ? "no-cache" : undefined
		return fetch(INDEX, { cache }).then(async indexRes => {
			if (indexRes.status >= 400) {
				this.setState({
					loading: false,
					error: indexRes.status !== 404 ? indexRes.statusText : null,
				})
				return
			}
			const indexData = await indexRes.json()
			const packNames = indexData.packs ?? []
			HOMESERVER_URL = indexData.homeserver_url || HOMESERVER_URL
			HOMESERVER_ANIMATED_URL = indexData.homeserver_animated_url || HOMESERVER_URL
			if (indexData.giphy_api_key !== undefined) {
				setGiphyAPIKey(indexData.giphy_api_key, indexData.giphy_mxc_prefix)
			}
			// TODO only load pack metadata when scrolled into view?
			const fetchedPacks = await Promise.all(packNames.map(async packFile => {
				let packRes
				if (packFile.startsWith("https://") || packFile.startsWith("http://")) {
					packRes = await fetch(packFile, {cache})
				} else {
					packRes = await fetch(`${PACKS_BASE_URL}/${packFile}`, {cache})
				}
				return await packRes.json()
			}))
			this.setState({
				packs: fetchedPacks,
				loading: false,
			})
			this.populateStickersByID(fetchedPacks)
			this.refreshFrequentlyUsed(fetchedPacks)
			return fetchedPacks
		}, error => this.setState({ loading: false, error }))
	}

	componentDidMount() {
		document.documentElement.style.setProperty("--stickers-per-row", this.state.stickersPerRow.toString())

		this._loadPacks()
		this._initializeStickersByID(frequent.getFromCache())

		this.imageObserver = new IntersectionObserver(this.observeImageIntersections, { rootMargin: "100px" })
		this.sectionObserver = new IntersectionObserver(this.observeSectionIntersections)
	}

	observeImageIntersections(intersections) {
		for (const entry of intersections) {
			const img = entry.target.children.item(0)
			if (entry.isIntersecting) {
				img.setAttribute("src", img.getAttribute("data-src"))
				img.classList.add("visible")
			} else {
				img.removeAttribute("src")
				img.classList.remove("visible")
			}
		}
	}

	observeSectionIntersections(intersections) {
		const navWidth = this.navRef.getBoundingClientRect().width
		let minX = 0, maxX = navWidth
		let minXElem = null
		let maxXElem = null
		for (const entry of intersections) {
			const packID = entry.target.getAttribute("data-pack-id")
			if (!packID) {
				continue
			}
			const navElement = document.getElementById(`nav-${packID}`)
			if (!navElement) {
				continue
			}
			if (entry.isIntersecting) {
				navElement.classList.add("visible")
				const bb = navElement.getBoundingClientRect()
				if (bb.x < minX) {
					minX = bb.x
					minXElem = navElement
				} else if (bb.right > maxX) {
					maxX = bb.right
					maxXElem = navElement
				}
			} else {
				navElement.classList.remove("visible")
			}
		}
		if (minXElem !== null) {
			minXElem.scrollIntoView({inline: "start"})
		} else if (maxXElem !== null) {
			maxXElem.scrollIntoView({inline: "end"})
		}
	}

	componentDidUpdate() {
		if (this.packListRef === null) {
			return
		}
		for (const elem of this.packListRef.getElementsByClassName("sticker")) {
			this.imageObserver.observe(elem)
		}
		for (const elem of this.packListRef.children) {
			this.sectionObserver.observe(elem)
		}
	}

	componentWillUnmount() {
		this.imageObserver.disconnect()
		this.sectionObserver.disconnect()
	}

	sendSticker(sticker) {
		frequent.add(sticker.id)
		this.updateFrequentlyUsed()
		this.resetSearch()
		widgetAPI.sendSticker(sticker)
	}

	navScroll(evt) {
		this.navRef.scrollLeft += evt.deltaY
	}

	render() {
		const theme = `theme-${this.state.theme}`

		const filterActive = !!this.state.filteredPacks
		const packs = filterActive ? this.state.filteredPacks : [this.state.frequentlyUsed, ...this.state.packs]
		const noPacksForSearch = filterActive && packs.length === 0

		if (this.state.loading) {
			return html`
				<main class="spinner ${theme}">
					<${Spinner} size=${80} green/>
				</main>
			`
		} else if (this.state.error) {
			return html`
				<main class="error ${theme}">
					<h1>Failed to load packs</h1>
					<p>${this.state.error}</p>
				</main>
			`
		} else if (this.state.packs.length === 0) {
			return html`
				<main class="empty ${theme}"><h1>No packs found 😿</h1></main>
			`
		}

		const onClickOverride = this.state.viewingGifs
			? (evt, packID) => {
				evt.preventDefault()
				this.setState({viewingGifs: false}, () => {
					scrollToSection(null, packID)
				})
			} : null
		const switchToGiphy = () => this.setState({viewingGifs: true, filtering: defaultState.filtering})

		return html`
			<main class="has-content ${theme}">
				<nav onWheel=${this.navScroll} ref=${elem => this.navRef = elem}>
					${giphyIsEnabled() && html`
						<${NavBarItem} pack=${{id: "giphy", title: "GIPHY"}} iconOverride="giphy" onClickOverride=${switchToGiphy} extraClass=${this.state.viewingGifs ? "visible" : ""}/>
					`}
					<${NavBarItem} pack=${this.state.frequentlyUsed} iconOverride="recent" onClickOverride=${onClickOverride}/>
					${this.state.packs.map(pack => html`<${NavBarItem} id=${pack.id} pack=${pack} onClickOverride=${onClickOverride}/>`)}
					<${NavBarItem} pack=${{id: "settings", title: "Settings"}} iconOverride="settings" onClickOverride=${onClickOverride}/>
				</nav>

				${this.state.viewingGifs ? html`
					<${GiphySearchTab}/>
				` : html`
					<${SearchBox}
						onInput=${this.searchStickers}
						value=${this.state.searchTerm}
						onReset=${this.resetSearch}
					/>
					<div class="pack-list ${isMobileSafari ? "ios-safari-hack" : ""}" ref=${(elem) => (this.packListRef = elem)}>
						${noPacksForSearch ? html`<div class="search-empty"><h1>No stickers match your search</h1></div>` : null}
						${packs.map(pack => html`<${Pack}
							id=${pack.id}
							pack=${pack}
							send=${this.sendSticker}
							hideGifIndicator=${this.state.userSettings.hideGifIndicator}
						/>`)}
						<${Settings} app=${this}/>
					</div>
				`}
			</main>`
	}
}

const Settings = ({app}) => html`
	<section class="stickerpack settings" id="pack-settings" data-pack-id="settings">
		<h1>Settings</h1>
		<div class="settings-list">
			<button onClick=${app.reloadPacks}>Reload</button>
			<button onClick=${app.clearFrequentlyUsed}>Clear frequently used</button>
			<div>
				<label for="stickers-per-row">Stickers per row: ${app.state.stickersPerRow}</label>
				<input type="range" min=2 max=10 id="stickers-per-row" id="stickers-per-row"
					value=${app.state.stickersPerRow}
					onInput=${evt => app.setStickersPerRow(evt.target.value)}/>
			</div>
			<div>
				<label for="theme">Theme: </label>
				<select name="theme" id="theme" onChange=${evt => app.setTheme(evt.target.value)}>
					<option value="default">Default</option>
					<option value="light">Light</option>
					<option value="dark">Dark</option>
					<option value="black">Black</option>
				</select>
			</div>
			<div>
				<label for="hide-gif-indicator">Hide GIF indicator:</label>
				<input
					type="checkbox"
					id="hide-gif-indicator"
					checked=${app.state.userSettings.hideGifIndicator}
					onChange=${evt => app.setDisplayGifIndicator(evt.target.checked)}
				/>
			</div>
		</div>
	</section>
`

// By default we just let the browser handle scrolling to sections, but webviews on Element iOS
// open the link in the browser instead of just scrolling there, so we need to scroll manually:
const scrollToSection = (evt, id) => {
	const pack = document.getElementById(`pack-${id}`)
	if (pack) {
		pack.scrollIntoView({block: "start", behavior: "instant"})
	}
	evt?.preventDefault()
}

const NavBarItem = ({pack, iconOverride = null, onClickOverride = null, extraClass = null}) => html`
	<a href="#pack-${pack.id}" id="nav-${pack.id}" data-pack-id=${pack.id} title=${pack.title} class="${extraClass}"
	   onClick=${onClickOverride ? (evt => onClickOverride(evt, pack.id)) : (isMobileSafari ? (evt => scrollToSection(evt, pack.id)) : undefined)}>
		<div class="sticker">
			${iconOverride ? html`
				<span class="icon icon-${iconOverride}"/>
			` : html`
				<img src=${makeThumbnailURL(pack.stickers[0])}
					alt=${pack.stickers[0].body} class="visible" />
			`}
		</div>
	</a>
`

const Pack = ({ pack, send, hideGifIndicator }) => html`
	<section class="stickerpack" id="pack-${pack.id}" data-pack-id=${pack.id}>
		<h1>${pack.title}</h1>
		<div class="sticker-list">
			${pack.stickers.map(sticker => html`
				<${Sticker} key=${sticker.id} sticker=${sticker} send=${send} hideGifIndicator=${hideGifIndicator} />
			`)}
		</div>
	</section>
`

const gifIcon = html`<div class="gif-indicator">GIF</div>`

const Sticker = ({ sticker, send, hideGifIndicator }) => html`
	<div class="sticker" onClick=${() => send(sticker)}>
		<img data-src=${makeThumbnailURL(sticker)} alt=${sticker.body} title=${sticker.body} />
		${sticker.info["mimetype"] === "image/gif" && !hideGifIndicator ? gifIcon : ''}
	</div>
`

render(html`<${App}/>`, document.body)
